@@ -5,18 +5,21 @@ Test database general API in available engines.
 
 Running Postgres test needs sufficient variables in environment like `PGUSER`.
 
+Running Postgres concurrent test with two connections needs at least `PGDBNAME2` in environment.
+
 ------------------------------------------------------------------------------
 This file is part of dblite - simple query interface for SQL databases.
 Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     20.11.2022
-@modified    06.12.2022
+@modified    27.03.2023
 ------------------------------------------------------------------------------
 """
 import collections
 import contextlib
 import copy
+import itertools
 import logging
 import os
 import tempfile
@@ -38,6 +41,12 @@ class TestAPI(unittest.TestCase):
         "postgres": ({}, {"maxconn": 2}),
     }
 
+    ## Engine-specific argument placeholder templates in raw SQL
+    ARGHOLDERS = {
+        "sqlite":   {"pos": "?",   "named": ":%(name)s"},
+        "postgres": {"pos": "%%s", "named": "%%(%(name)s)s"},
+    }
+
     ## Table columns as {table name: [{"name", "type"}]}
     TABLES = {
         "test": [{"name": "id",  "type": "INTEGER PRIMARY KEY"},
@@ -60,14 +69,15 @@ class TestAPI(unittest.TestCase):
         try: unittest.util._MAX_LENGTH = 100000
         except Exception: pass
         self._connections = collections.OrderedDict()  # {engine: (opts, kwargs)}
-        self._path = None  # Path to SQLite database
+        self._paths = []  # Paths to SQLite databases
 
 
     def setUp(self):
         """Creates engine connection options."""
         super(TestAPI, self).setUp()
-        with tempfile.NamedTemporaryFile(suffix=".sqlite") as f: self._path = f.name
-        self._connections["sqlite"] = (self._path, self.ENGINES["sqlite"][1])
+        for _ in range(2):
+            with tempfile.NamedTemporaryFile(suffix=".sqlite") as f: self._paths.append(f.name)
+        self._connections["sqlite"] = (self._paths[0], self.ENGINES["sqlite"][1])
 
         try: import psycopg2
         except ImportError:
@@ -85,8 +95,9 @@ class TestAPI(unittest.TestCase):
 
     def tearDown(self):
         """Deletes temoorary files and tables."""
-        try: os.remove(self._path)
-        except Exception: pass
+        for p in self._paths:
+            try: os.remove(p)
+            except Exception: pass
         try:
             opts, kwargs = self._connections["postgres"]
             with dblite.init(opts, "postgres", **kwargs) as db:
@@ -103,17 +114,91 @@ class TestAPI(unittest.TestCase):
                 if i: logger.info("-" * 60)
                 logger.info("Verifying dblite API for %s.", engine)
                 self.verify_general_api(opts, kwargs, engine)
-                self.verify_query_api(dblite)
+                self.verify_query_api(dblite, engine)
                 dblite.close()
                 with dblite.init() as db:
-                    self.verify_query_api(db)
+                    self.verify_query_api(db, engine)
                 with dblite.transaction() as tx:
-                    self.verify_query_api(tx)
+                    self.verify_query_api(tx, engine)
                 self.verify_query_args(dblite)
                 self.verify_transactions()
                 self.verify_exclusive_transactions()
 
 
+    def test_concurrent(self):
+        """Tests dblite API with multiple concurrent engines and databases."""
+        logger.info("Verifying dblite API with multiple concurrent engines.")
+        dbs = [dblite.init(p) for p in self._paths]
+        if "postgres" in self._connections:
+            opts, kwargs = self._connections["postgres"]
+            dbs.append(dblite.init(opts, engine="postgres", **kwargs))
+            if os.getenv("PGDBNAME2"):
+                opts2 = dict(opts, **{k[2:-1].lower(): v for k, v in os.environ.items()
+                                      if k.startswith("PG") and k.endswith("2")})
+                dbs.append(dblite.init(opts2, engine="postgres", **kwargs))
+        for db in dbs:
+            self.verify_query_api(db, db.ENGINE)
+        self.verify_concurrent(dbs)
+
+
+    def verify_concurrent(self, dbs):
+        """Tests dblite API with multiple concurrent databases."""
+        logger.info("Verifying %s concurrent databases with engines %s.",
+                    len(dbs), set(db.ENGINE for db in dbs))
+
+        TABLES = collections.defaultdict(set)  # {db: (table, )}
+        IDS = collections.defaultdict(dict)  # {db: {table: [id, ]}}
+
+        logger.debug("Verifying creating tables in concurrent connections.")
+        for table, cols in self.TABLES.items():
+            for i, db in enumerate(dbs):
+                for name in [table, "%s_%s" % (table, i)]:
+                    TABLES[db].add(name)
+                    db.executescript("CREATE TABLE %s (%s)" %
+                                     (name, ", ".join("%(name)s %(type)s" % c for c in cols)))
+                    self.assertFalse(db.fetchall(name),
+                                     "Unexpected value from %s.fetchall()." % label(db))
+
+        logger.debug("Verifying inserting data in concurrent connections.")
+        cycletables = itertools.cycle((d, t) for d, tt in TABLES.items() for t in tt)
+        for i in range(1, 3 * sum(map(len, TABLES.values())) + 1):
+            db, table = next(cycletables)
+            db.insert(table, id=i, val="val_%s" % i)
+            IDS[db].setdefault(table, []).append(i)
+
+        logger.debug("Verifying selecting in concurrent connections.")
+        for db in IDS:
+            for table, ids in IDS[db].items():
+                received = set(x["id"] for x in db.fetchall(table))
+                self.assertEqual(received, set(ids),
+                                 "Unexpected value from %s.fetchall()." % label(db))
+
+        logger.debug("Verifying selecting in concurrent transactions.")
+        txs = {db: db.transaction() for db in dbs}
+        for tx in txs.values(): tx.__enter__()
+        for db, tx in txs.items():
+            for table, ids in IDS[db].items():
+                received = set(x["id"] for x in tx.fetchall(table))
+                self.assertEqual(received, set(ids),
+                                 "Unexpected value from %s.fetchall()." % label(tx))
+        for tx in txs.values(): tx.close()
+
+        logger.debug("Verifying deleting in concurrent connections.")
+        while any(any(xx.values()) for xx in IDS.values()):
+            db, table, ids = next((d, t, vv) for d, xx in IDS.items() for t, vv in xx.items() if vv)
+            self.assertTrue(db.delete(table, id=ids.pop()),
+                            "Unexpected value from %s.delete()." % label(db))
+        for db, table in ((d, t) for d, tt in TABLES.items() for t in tt):
+            self.assertFalse(db.fetchone(table), "Unexpected value for %s.fetchone()." % label(db))
+
+        logger.debug("Verifying dropping in concurrent connections.")
+        for db, table in ((d, t) for d, tt in TABLES.items() for t in tt):
+            db.executescript("DROP TABLE %s" % table)
+            with self.assertRaises(Exception,
+                                   msg="Unexpected success for fetch after dropping table."):
+                db.fetchone(table)
+
+
     def verify_general_api(self, opts, kwargs, engine):
         """Verifies general module-level and Database-level functions."""
         logger.info("Verifying dblite general module and Database functions.")
@@ -128,11 +213,13 @@ class TestAPI(unittest.TestCase):
         tx = db.transaction()
         self.assertIsInstance(tx, dblite.api.Transaction,
                               "Unexpected value from dblite.transaction().")
+        self.assertEqual(tx.ENGINE, engine, "Unexpected value from %s.ENGINE." % label(tx))
         tx.close()
 
         logger.info("Verifying Database property decorators.")
         self.assertFalse(db.closed, "Unexpected value from %s.closed." % label(db))
         self.assertIsNotNone(db.cursor, "Unexpected value from %s.cursor." % label(db))
+        self.assertEqual(db.ENGINE, engine, "Unexpected value from %s.ENGINE." % label(db))
 
         logger.info("Verifying dblite.close().")
         self.assertFalse(db.closed, "Unexpected value from %s.closed." % label(db))
@@ -145,7 +232,7 @@ class TestAPI(unittest.TestCase):
             db.fetchone(next(iter(self.TABLES)))
 
 
-    def verify_query_api(self, obj):
+    def verify_query_api(self, obj, engine):
         """Verifies query functions."""
         logger.info("Verifying %s query functions.", label(obj))
         DATAS = copy.deepcopy(self.DATAS)
@@ -206,6 +293,28 @@ class TestAPI(unittest.TestCase):
             affected = obj.delete(table)
             self.assertGreater(affected, 1, "Unexpected value from %s.delete()." % obj)
 
+            logger.debug("Verifying %s.insertmany(%r).", label(obj), table)
+            res = obj.insertmany(table, DATAS[table])
+            self.assertIsInstance(res, list, "Unexpected value from %s.insertmany()." % obj)
+            rows = obj.fetchall(table)
+            self.assertEqual(rows, DATAS[table], "Unexpected value from %s.fetchall()." % obj)
+            obj.delete(table)
+
+            logger.debug("Verifying %s.insertmany(%r) with keyword overrides.", label(obj), table)
+            res = obj.insertmany(table, DATAS[table], val="common for all")
+            rows = obj.fetchall(table)
+            self.assertEqual(rows, [dict(x, val="common for all") for x in DATAS[table]],
+                             "Unexpected value from %s.fetchall()." % obj)
+
+            logger.debug("Verifying %s.executemany(%r).", label(obj), table)
+            for cls in (dict, list):
+                tpl = self.ARGHOLDERS[engine]["named" if cls is dict else "pos"]
+                sql = "DELETE FROM %s WHERE id = %s" % (table, tpl % dict(name="id"))
+                args = ({"id": x["id"]} if cls is dict else [x["id"]] for x in DATAS[table])
+                obj.executemany(sql, args)
+                self.assertFalse(obj.fetchall(table), "Unexpected value from %s.fetchall()." % obj)
+                obj.insertmany(table, DATAS[table])
+
             logger.debug("Verifying %s.executescript().", label(obj))
             obj.executescript("DROP TABLE %s" % table)
             with self.assertRaises(Exception,
